#include <assert.h>
#include <errno.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <glib-object.h>
#include <xorcom/licensing.h>
#include "twinstar-dbus.h"
#include "twinstar-license.h"
#include "twinstar-cpg-glib.h"
#include "twinstar-corosync.h"
#include "twinstar-astribanks.h"
#include "twinstar-run.h"
#include "debug.h"

G_DEFINE_TYPE(TwinstarCorosync, twinstar_corosync, G_TYPE_OBJECT);

#define TWINSTAR_COROSYNC_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), TWINSTAR_TYPE_COROSYNC, TwinstarCorosyncPrivate))

#define	NO_PRIORITY		0
#define	ELECTION_TIME		500
#define	SYNC_INTERVAL		1	/* seconds */
#define	SYNC_PRIORITY_FACTOR	10

#define	LICENSE_VALIDATION_INTERVAL_SHORT		(10)		/* seconds */
#define	LICENSE_VALIDATION_INTERVAL_LONG		(60 * 60)	/* seconds */


struct node_desc {
	uint32_t nodeid;
	int32_t priority;	/* to be master. higher is better */
	uint32_t sync_counter;
} __attribute__((packed));

struct TwinstarCorosyncPrivate {
	TwinstarDBus *twdbus;
	TwinstarLicense *twlicense;
	TwinstarGCpg *twgcpg;
	TwinstarRun *twrun;
	TwinstarAstribanks *twab;
	guint election_timer;
	guint sync_timer;
	gboolean in_transition;
	gboolean local_has_valid_license;
	struct node_desc local;
	struct node_desc best_proposed;
	struct node_desc current_master;
};

/*
 * Message handlers definitions
 */

struct twinstar_message_header {
	uint8_t	mtype;
	uint8_t	reserved[3];
};

struct twinstar_message {
	struct twinstar_message_header header;
	char mdata[0];
} __attribute__((packed));

#define	TWINSTAR_MESSAGE_STRUCT(t, ...) \
	struct twinstar_message_ ## t

#define	DEFINE_TWINSTAR_MESSAGE(t, ...) \
	TWINSTAR_MESSAGE_STRUCT(t) { \
		struct twinstar_message_header header; \
		struct t ## _payload { \
		__VA_ARGS__ \
		} payload; \
	} __attribute__((packed))

#define	TWINSTAR_MESSAGE_HANDLER_NAME(n)	(message_handler_ ## n)
#define	TWINSTAR_MESSAGE_ENTRY(mtype, htype)	[mtype]	= TWINSTAR_MESSAGE_HANDLER_NAME(htype)

DEFINE_TWINSTAR_MESSAGE(text, \
		char text[0];	/* variable length */
	);

DEFINE_TWINSTAR_MESSAGE(keep_alive, \
	);

DEFINE_TWINSTAR_MESSAGE(sync, \
		struct node_desc node;
	);

DEFINE_TWINSTAR_MESSAGE(master_propose, \
		struct node_desc proposal;
		uint32_t current_master_nodeid;
	);

DEFINE_TWINSTAR_MESSAGE(master_takeover, \
		struct node_desc node;
	);

DEFINE_TWINSTAR_MESSAGE(frozen_master, \
		uint32_t nodeid;
	);

DEFINE_TWINSTAR_MESSAGE(license_ok, \
		gboolean license_ok;
	);

DEFINE_TWINSTAR_MESSAGE(master_election_start, \
	);

DEFINE_TWINSTAR_MESSAGE(astribank_notify, \
		uint32_t count;
		gboolean added;
		char hardware_id[40];
		char location[40];
		char devpath[0];	/* variable length */
	);

#define	ALLOCA_TWINSTAR_MESSAGE(mt, t, extra, ...) \
		({ \
			TWINSTAR_MESSAGE_STRUCT(t) *p; \
			int len = sizeof(TWINSTAR_MESSAGE_STRUCT(t)) + (extra); \
			p = alloca(len); \
			memset(p, 0, len); \
			p->header.mtype = (mt); \
			p->header.reserved[0] = 0; \
			p->header.reserved[1] = 0; \
			p->header.reserved[2] = 0; \
			p; \
		})

/*
 * General Methods
 */
static int send_master_election_start(TwinstarCorosync *twcorosync);
static int frozen_master_was_set(void *context, uint32_t nodeid);
static int send_license_update(TwinstarCorosync *twcorosync);

static void clear_node(struct node_desc *node)
{
	node->nodeid = UNKNOWN_NODEID;
	node->priority = NO_PRIORITY;
	node->sync_counter = 0;
}

uint32_t twinstar_corosync_local_nodeid(const void *context)
{
	const TwinstarCorosync *twcorosync = context;
	const TwinstarCorosyncPrivate *priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	return twinstar_glib_local_nodeid(priv->twgcpg);
}

int send_master_propose(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(master_propose) *msg;
	int ret;

	msg = ALLOCA_TWINSTAR_MESSAGE('P', master_propose, 0);
	if (!msg)
		return -ENOMEM;
	msg->payload.proposal = priv->local;
	msg->payload.current_master_nodeid = priv->current_master.nodeid;
	debug(TWINSTAR, "nodeid = %u priority = %d sync_counter = %u current_master = %u\n",
		priv->local.nodeid,
		priv->local.priority,
		priv->local.sync_counter,
		priv->current_master.nodeid);
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return ret;
}

static int send_master_takeover(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(master_takeover) *msg;
	int ret;

	debug(TWINSTAR, "nodeid = %u priority = %d\n",
		priv->local.nodeid, priv->local.priority);
	msg = ALLOCA_TWINSTAR_MESSAGE('M', master_takeover, 0);
	if (!msg)
		return -ENOMEM;
	msg->payload.node = priv->local;
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return ret;
}

static void set_master(TwinstarCorosync *twcorosync, const struct node_desc *node)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarDBus *twdbus = priv->twdbus;
	gboolean got_master = (node->nodeid == priv->local.nodeid);

	priv->in_transition = TRUE;
	if (got_master) {
		info("MASTER: (nodeid = %u priority = %d sync_counter = %u) -- local sync_counter = %u\n",
			node->nodeid,
			node->priority,
			node->sync_counter,
			priv->local.sync_counter);
	} else {
		info("SLAVE: (master nodeid = %u priority = %d sync_counter = %u) -- local sync_counter = %u\n",
			node->nodeid,
			node->priority,
			node->sync_counter,
			priv->local.sync_counter);
	}
	priv->current_master = *node;
	twinstar_notify_new_master(twdbus, node->nodeid);
	if (priv->twrun) {
		gboolean result = twinstar_run_start(priv->twrun, got_master);
		debug(TWINSTAR, "Started agents: status = %d\n", result);
	}
}

static gboolean election_timed_out(gpointer user_data)
{
	TwinstarCorosync *twcorosync = user_data;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	if (! VALID_NODEID(priv->current_master.nodeid))
		debug(TWINSTAR, "previous master -- Unknown\n");
	else
		debug(TWINSTAR, "previous master -- %u\n",
			priv->current_master.nodeid);
	if (! VALID_NODEID(priv->best_proposed.nodeid)) {
		info("Failed to find a master -- restarting election...\n");
		send_master_election_start(twcorosync);
	} else {
		debug(TWINSTAR, "Found best master: nodeid = %u priority = %d\n",
			priv->best_proposed.nodeid, priv->best_proposed.priority);
		if (priv->best_proposed.nodeid == priv->current_master.nodeid) {
			info("ELECTION-FINISHED: Master stays %u [priority %d]\n",
				priv->best_proposed.nodeid,
				priv->best_proposed.priority);
		} else if (priv->best_proposed.nodeid == priv->local.nodeid) {
			info("ELECTION-FINISHED: Master should be %u (that's me -- taking over) [priority %d]\n",
				priv->best_proposed.nodeid,
				priv->best_proposed.priority);
			send_master_takeover(twcorosync);
		} else {
			info("ELECTION-FINISHED: Master should be %u (it should take over) [priority %d]\n",
				priv->best_proposed.nodeid,
				priv->best_proposed.priority);
			set_master(twcorosync, &priv->best_proposed);
		}
	}
	return FALSE;
}

static void do_master_election_start(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	if (priv->election_timer) {
		g_source_remove(priv->election_timer);
		priv->election_timer = 0;
	}
	clear_node(&priv->best_proposed);
	priv->election_timer = g_timeout_add(ELECTION_TIME, election_timed_out, twcorosync);
}

int send_master_election_start(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(master_election_start) *msg;
	int ret;

	priv->local.sync_counter += SYNC_PRIORITY_FACTOR * priv->local.priority;
	msg = ALLOCA_TWINSTAR_MESSAGE('1', master_election_start, 0);
	if (!msg)
		return -ENOMEM;
	debug(TWINSTAR, "START MASTER ELECTION (local nodeid = %u priority = %d)\n",
		priv->local.nodeid, priv->local.priority);
	do_master_election_start(twcorosync);
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return ret;
}

#ifdef	DEBUG
static void message_dump(const char *msg, size_t msg_len)
{
	int i;

	for (i = 0; i < msg_len; i++)
		debug(TWINSTAR, "\t[%02d] = 0x%X\n", i, (uint8_t)(msg[i]));
}
#endif


/*
 * Message handlers
 */

typedef int (*twcorosync_message_handler_t)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len);

int TWINSTAR_MESSAGE_HANDLER_NAME(text)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarDBus *twdbus = priv->twdbus;
	TWINSTAR_MESSAGE_STRUCT(text) *p = (TWINSTAR_MESSAGE_STRUCT(text) *)msg;
	char *text;

	text = p->payload.text;
	debug(TWINSTAR, "'%s' (current_master=%u)\n", text, priv->current_master.nodeid);
	twinstar_notify_got_message(twdbus, text);
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(keep_alive)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarDBus *twdbus = priv->twdbus;
	//TWINSTAR_MESSAGE_STRUCT(keep_alive) *p = (TWINSTAR_MESSAGE_STRUCT(keep_alive) *)msg;

	debug(TWINSTAR, "from nodeid=%u\n", nodeid);
	twinstar_notify_keep_alive(twdbus, nodeid);
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(sync)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TWINSTAR_MESSAGE_STRUCT(sync) *p = (TWINSTAR_MESSAGE_STRUCT(sync) *)msg;

	if (priv->current_master.nodeid == p->payload.node.nodeid) {
		if (priv->current_master.nodeid != priv->local.nodeid) {
			/* Sync to master */
			debug(TWINSTAR, "from Master (nodeid %u sync_counter %u)\n",
				p->payload.node.nodeid,
				p->payload.node.sync_counter);
			priv->local.sync_counter = p->payload.node.sync_counter;
		}
	} else {
		debug(TWINSTAR, "retain Slave (nodeid %u sync_counter %u)\n",
			priv->local.nodeid,
			priv->local.sync_counter);
	}
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(master_propose)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TWINSTAR_MESSAGE_STRUCT(master_propose) *p = (TWINSTAR_MESSAGE_STRUCT(master_propose) *)msg;

	debug(TWINSTAR, "From %u: nodeid = %u priority = %d sync_counter = %u current_master_nodeid = %u (my_priority = %d)\n",
		nodeid,
		p->payload.proposal.nodeid,
		p->payload.proposal.priority,
		p->payload.proposal.sync_counter,
		p->payload.current_master_nodeid,
		priv->local.priority);
	if (VALID_NODEID(p->payload.current_master_nodeid) &&
		VALID_NODEID(priv->current_master.nodeid) &&
		priv->current_master.nodeid != p->payload.current_master_nodeid) {
		err("From %u: SPLIT BRAIN RECONNECTED (current %d, got %d) -- Aborting\n",
			nodeid,
			priv->current_master.nodeid,
			p->payload.current_master_nodeid);
		abort();
	}
	/* First proposal? */
	if (! VALID_NODEID(priv->best_proposed.nodeid)) {
		priv->best_proposed = p->payload.proposal;
		debug(TWINSTAR, "From %u: PROPOSAL FROM %u: FIRST: priority %d sync_counter %u\n",
					nodeid,
					priv->best_proposed.nodeid,
					priv->best_proposed.priority,
					priv->best_proposed.sync_counter);
	} else {
		if (priv->best_proposed.sync_counter < p->payload.proposal.sync_counter) {
			/* Got better proposal */
			priv->best_proposed = p->payload.proposal;
			debug(TWINSTAR, "From %u: PROPOSAL FROM %u: BETTER SYNC COUNTER: priority %d sync_counter %u\n",
						nodeid,
						priv->best_proposed.nodeid,
						priv->best_proposed.priority,
						priv->best_proposed.sync_counter);
		}
	}
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(master_election_start)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	debug(TWINSTAR, "From %u: \n", nodeid);
	do_master_election_start(twcorosync);
	send_master_propose(twcorosync);
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(master_takeover)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TWINSTAR_MESSAGE_STRUCT(master_takeover) m;

	memcpy((void *)&m, msg, sizeof(m));
	debug(TWINSTAR, "From %u: Master is now %u priority = %d sync_counter = %u\n",
		nodeid,
		m.payload.node.nodeid,
		m.payload.node.priority,
		m.payload.node.sync_counter);
	set_master(twcorosync, &m.payload.node);
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(frozen_master)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TWINSTAR_MESSAGE_STRUCT(frozen_master) *p = (TWINSTAR_MESSAGE_STRUCT(frozen_master) *)msg;
	uint32_t frozen_master = p->payload.nodeid;

	debug(TWINSTAR, "From %u: Got frozen Master = %u\n", nodeid, frozen_master);
	twinstar_notify_frozen_master(priv->twdbus, frozen_master);
	if (VALID_NODEID(frozen_master)) {
		const char *msg = (VALID_NODEID(priv->current_master.nodeid))
					? "already"
					: "will be";
		if (frozen_master == priv->local.nodeid) {
			info("From %u: FROZEN MASTER %u -- %s master\n", nodeid, frozen_master, msg);
			if (! VALID_NODEID(priv->current_master.nodeid)) {
				/* A restart/reboot after freeze */
				send_master_takeover(twcorosync);
			}
		} else {
			info("From %u: FROZEN MASTER %u -- %s slave\n", nodeid, frozen_master, msg);
		}
	} else {
		info("From %u: UNFROZEN master -- start election\n", nodeid);
		send_master_election_start(twcorosync);
	}
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(license_ok)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TWINSTAR_MESSAGE_STRUCT(license_ok) *p = (TWINSTAR_MESSAGE_STRUCT(license_ok) *)msg;
	gboolean license_ok = p->payload.license_ok;

	debug(TWINSTAR, "From %u: license_ok = %s\n",
		nodeid, (license_ok) ? "YES" : "NO");
	twinstar_license_update(priv->twlicense, nodeid, license_ok);
	return 0;
}

int TWINSTAR_MESSAGE_HANDLER_NAME(astribank_notify)(TwinstarCorosync *twcorosync, uint32_t nodeid, uint8_t mtype, const char *msg, size_t msg_len)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TWINSTAR_MESSAGE_STRUCT(astribank_notify) *p = (TWINSTAR_MESSAGE_STRUCT(astribank_notify) *)msg;

	info("From %u: now %d Astribanks: %s [%s] @%s (%s)\n",
		nodeid,
		p->payload.count,
		(p->payload.added) ? "ADDED" : "REMOVED",
		p->payload.hardware_id,
		p->payload.location,
		p->payload.devpath);
	twinstar_notify_astribank_hotplug(priv->twdbus,
		nodeid,
		p->payload.count,
		p->payload.added,
		p->payload.hardware_id,
		p->payload.location,
		p->payload.devpath);
	return 0;
}

static twcorosync_message_handler_t message_handlers[256] = {
	TWINSTAR_MESSAGE_ENTRY('T', text),
	TWINSTAR_MESSAGE_ENTRY('K', keep_alive),
	TWINSTAR_MESSAGE_ENTRY('s', sync),
	TWINSTAR_MESSAGE_ENTRY('1', master_election_start),
	TWINSTAR_MESSAGE_ENTRY('P', master_propose),
	TWINSTAR_MESSAGE_ENTRY('M', master_takeover),
	TWINSTAR_MESSAGE_ENTRY('F', frozen_master),
	TWINSTAR_MESSAGE_ENTRY('L', license_ok),
	TWINSTAR_MESSAGE_ENTRY('A', astribank_notify),
};

/*
 * Methods from COROSYNC CPG to DBUS
 */

void handle_joined(void *context, uint32_t nodeid, size_t node_count)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarDBus *twdbus = priv->twdbus;
	uint32_t frozen_master;

	debug(TWINSTAR, "From %u: JOINED: %zd nodes\n",
		nodeid, node_count);
	frozen_master = twinstar_get_frozen_master(twdbus);
	if (VALID_NODEID(frozen_master)) {
		info("From %u: JOINED -- no master election (FROZEN to %u)\n", nodeid, frozen_master);
	} else {
		send_master_election_start(twcorosync);
	}
	twinstar_license_joined(priv->twlicense, nodeid);
	send_license_update(twcorosync);
	twinstar_notify_joined(twdbus, nodeid, node_count);
	/* Update everybody, so joined node is aware of frozen state */
	frozen_master_was_set(twcorosync, frozen_master);
}

void handle_left(void *context, uint32_t nodeid, size_t node_count)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarDBus *twdbus = priv->twdbus;
	uint32_t frozen_master;

	debug(TWINSTAR, "From %u: LEFT %zd nodes\n", nodeid, node_count);
	frozen_master = twinstar_get_frozen_master(twdbus);
	if (VALID_NODEID(frozen_master)) {
		info("From %u: LEFT -- no master election (FROZEN to %u)\n", nodeid, frozen_master);
	} else {
		send_master_election_start(twcorosync);
	}
	twinstar_license_left(priv->twlicense, nodeid);
	twinstar_notify_left(twdbus, nodeid, node_count);
}

void handle_got_message(void *context, uint32_t nodeid, void *msg, size_t msg_len)
{
	TwinstarCorosync *twcorosync = context;
	const struct twinstar_message *tw_msg = msg;
	uint8_t mtype = tw_msg->header.mtype;
	twcorosync_message_handler_t mhandler = message_handlers[mtype];

	if (!mhandler) {
		err("From %u: Message mtype=%u(%c) without handler. Discarding (%zu bytes)\n",
			nodeid, mtype, mtype, msg_len);
		return;
	}
	//debug(TWINSTAR, "mtype=%u(%c) -- %u bytes\n", mtype, mtype, msg_len);
	//message_dump(msg, msg_len);
	mhandler(twcorosync, nodeid, mtype, msg, msg_len);
}

static struct twinstar_cpg_notify_ops cpg_ops = {
	.joined		= handle_joined,
	.left		= handle_left,
	.got_message	= handle_got_message,
};

/*
 * Methods from DBUS to COROSYNC CPG
 */
static int handle_text_message(void *context, const char *text)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(text) *msg;
	size_t text_len = strlen(text) + 1;	/* copy the '\0' */

	debug(TWINSTAR, "'%s'\n", text);
	msg = ALLOCA_TWINSTAR_MESSAGE('T', text, text_len);
	if (!msg)
		return -ENOMEM;
	memcpy(msg->payload.text, text, text_len);
	return twcpg_glib_send_message(twgcpg, msg, sizeof(msg->header) + text_len);
}

static int handle_keep_alive(void *context, uint32_t nodeid)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(keep_alive) *msg;

	debug(TWINSTAR, "nodeid=%u\n", nodeid);
	msg = ALLOCA_TWINSTAR_MESSAGE('K', keep_alive, 0);
	if (!msg)
		return -ENOMEM;
	return twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
}

static int preferred_master_was_set(void *context, gboolean preferred_master)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	uint32_t frozen_master;

	debug(TWINSTAR, "%u\n", preferred_master);
	if (twinstar_get_preferred_master(priv->twdbus))
		priv->local.priority = 1;
	else
		priv->local.priority = 0;
	/* Re-election */
	frozen_master = twinstar_get_frozen_master(priv->twdbus);
	if (VALID_NODEID(frozen_master)) {
		info("PREFERRED MASTER(%u) -- no master election (FROZEN to %u)\n",
			preferred_master, frozen_master);
	} else {
		send_master_election_start(twcorosync);
	}
	return 0;
}

static int frozen_master_was_set(void *context, uint32_t nodeid)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(frozen_master) *msg;
	int ret;

	debug(TWINSTAR, "nodeid = %u frozen_master = %u\n",
		priv->local.nodeid, nodeid);
	msg = ALLOCA_TWINSTAR_MESSAGE('F', frozen_master, 0);
	if (!msg)
		return -ENOMEM;
	msg->payload.nodeid = nodeid;
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return ret;
}

static int send_license_update(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(license_ok) *msg;
	int ret;

	debug(TWINSTAR, "nodeid = %u license_ok = %s\n",
		priv->local.nodeid, (priv->local_has_valid_license) ? "YES" : "NO");
	msg = ALLOCA_TWINSTAR_MESSAGE('L', license_ok, 0);
	if (!msg)
		return -ENOMEM;
	msg->payload.license_ok = priv->local_has_valid_license;
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return ret;
}

static int do_take_over(void *context)
{
	TwinstarCorosync *twcorosync = context;

	send_master_takeover(twcorosync);
	return 0;
}

static int do_astribank_hotplug(void *context, gboolean added,
			const char *devpath,
			const char *hardware_id,
			const char *location)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	debug(TWINSTAR, "%s: %s [%s] @%s %s\n", __func__, (added) ? "ADDED" : "REMOVED", hardware_id, location, devpath);
	/* Update Astribanks counter */
	twinstar_astribank_event(priv->twab, added, devpath);
	return 0;
}

static int do_astribanks_count(void *context)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	int ab_count;

	ab_count = twinstar_astribank_count(priv->twab);
	debug(TWINSTAR, "got %d\n", ab_count);
	return ab_count;
}

struct twinstar_dbus_notify_ops dbus_ops = {
	.send_message	= handle_text_message,
	.keep_alive	= handle_keep_alive,
	.preferred_master_was_set = preferred_master_was_set,
	.frozen_master_was_set = frozen_master_was_set,
	.take_over = do_take_over,
	.astribank_hotplug = do_astribank_hotplug,
	.astribanks_count = do_astribanks_count,
};

/*
 * TwinstarRun agents
 */

static void run_step_start(void *context, const char *path, gboolean is_master, pid_t pid)
{
	const TwinstarCorosync *twcorosync = context;
	const TwinstarCorosyncPrivate *priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	info("---------- started %s (is_master=%d) (pid = %d)\n", path, is_master, pid);
	twinstar_notify_runstep_begin(priv->twdbus, path, (is_master) ? "start" : "stop");
}

static void run_step_end(void *context, const char *path, gboolean is_master, pid_t pid, int status)
{
	const TwinstarCorosync *twcorosync = context;
	const TwinstarCorosyncPrivate *priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	info("---------- ended %s (is_master=%d) (pid = %d, status = %d)\n",
		path, is_master, pid, status);
	if (status) {
		char tmpbuf[BUFSIZ];
		int len = 0;

		len += snprintf(tmpbuf + len, sizeof(tmpbuf) - len,
			"%s(pid = %d): %s %s -- ",
			__func__, pid, path, (is_master) ? "master" : "slave");
		if (WIFEXITED(status)) {
			len += snprintf(tmpbuf + len, sizeof(tmpbuf) - len,
				"exited, status %d", WEXITSTATUS(status));
		} else if (WIFSIGNALED(status)) {
			len += snprintf(tmpbuf + len, sizeof(tmpbuf) - len,
				"killed by signal %d", WTERMSIG(status));
		} else if (WIFSTOPPED(status)) {
			len += snprintf(tmpbuf + len, sizeof(tmpbuf) - len,
				"stopped by signal %d", WSTOPSIG(status));
		} else if (WIFCONTINUED(status)) {
			len += snprintf(tmpbuf + len, sizeof(tmpbuf) - len,
				"continued");
		}
		err("%s\n", tmpbuf);
	}
	twinstar_notify_runstep_end(priv->twdbus, path, (is_master) ? "start" : "stop", status);
}

static void run_steps_done(void *context, int errors)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate *priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	info("---------- run steps done (%d errors)\n", errors);
	if (errors)
		err("%s: done with %d errors\n", __func__, errors);
	twinstar_notify_runsteps_done(priv->twdbus, errors);
	priv->in_transition = FALSE;
	twinstar_license_periodic_validation(priv->twlicense, LICENSE_VALIDATION_INTERVAL_SHORT);
}

static struct twinstar_run_notify_ops agents_notifications = {
	.run_step_start = run_step_start,
	.run_step_end = run_step_end,
	.run_steps_done = run_steps_done,
};

gboolean twinstar_corosync_set_agents(TwinstarCorosync *corosync, const char *agents_dir)
{
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(corosync);

	/* First cleanup */
	if (priv->twrun) {
		err("%s: cannot be run twice\n", __func__);
		return FALSE;
	}
	/* New jobs? */
	if (agents_dir) {
		priv->twrun = twinstar_run_new(agents_dir, &agents_notifications, corosync);
		return (priv->twrun) ? TRUE : FALSE;
	}
	return TRUE;
}


/*
 * License stuff
 */

/* license callback */
int cluster_license_notify(void *context, gboolean license_ok, gboolean changed)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	if (changed) {
		info("TWINSTAR CLUSTER LICENSE CHANGE: now %s\n",
			(license_ok) ? "VALID" : "INVALID");
		twinstar_notify_license(priv->twdbus, license_ok);
		if (license_ok) {
			twinstar_license_periodic_validation(priv->twlicense, LICENSE_VALIDATION_INTERVAL_LONG);
		}
	}
	if (! license_ok) {
		uint32_t frozen_master = twinstar_get_frozen_master(priv->twdbus);

		/* Already frozen? */
		if (!VALID_NODEID(frozen_master)) {
			uint32_t wanted_frozen = priv->current_master.nodeid;

			/* Safe to freeze now? */
			if (VALID_NODEID(wanted_frozen) && ! priv->in_transition) {
				warn("===> TWINSTAR CLUSTER LICENSE INVALID: freezing failover (master == %u)\n",
					wanted_frozen);
				frozen_master_was_set(twcorosync, wanted_frozen);
			} else {
				debug(TWINSTAR, "===> TWINSTAR CLUSTER LICENSE INVALID: not freezing yet (master == %u, in_transition == %d)\n",
					wanted_frozen, priv->in_transition);
			}
		}
	}
	return 0;
}

int local_license_notify(void *context, gboolean license_ok, gboolean changed)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	if (changed) {
		info("LOCAL LICENSE CHANGED TO: %s\n",
			(license_ok) ? "VALID" : "INVALID");
		priv->local_has_valid_license = license_ok;
	}
	/* Must always update so cluster-license is recalculated and we get callbacks */
	send_license_update(twcorosync);
	return 0;
}

static struct twinstar_license_notify_ops license_notify_ops = {
	.cluster_license_notify	= cluster_license_notify,
	.local_license_notify	= local_license_notify,
};

/*
 * TwinstarAstribanks handling
 */
static int twab_astribank_notify(int count, const struct astribank_info *ab, int added, const char *devpath, void *context)
{
	TwinstarCorosync *twcorosync = context;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(astribank_notify) *msg;
	int devpath_len = strlen(devpath);

	debug(TWINSTAR, "total %d: %s [%s] @%s (%s)\n",
		count,
		(added) ? "ADDED" : "REMOVED",
		ab->hardware_id,
		ab->location,
		devpath);
	msg = ALLOCA_TWINSTAR_MESSAGE('A', astribank_notify, devpath_len + 1);
	if (!msg)
		return -ENOMEM;
	msg->payload.count = count;
	msg->payload.added = added;
	snprintf(msg->payload.hardware_id, sizeof(msg->payload.hardware_id),
		"%s", ab->hardware_id);
	snprintf(msg->payload.location, sizeof(msg->payload.location),
		"%s", ab->location);
	/*
	 * We don't use snprintf(), since payload.devpth is 0-length
	 * If compiled with -Wp,-D_FORTIFY_SOURCE=2 it will add a runtime
	 * check [snprintf_chk()] that would blow our program
	 * for suspected "buffer-overflow".
	 */
	memcpy(msg->payload.devpath, devpath, devpath_len);
	msg->payload.devpath[devpath_len] = '\0';
	return twcpg_glib_send_message(twgcpg, msg, sizeof(*msg) + devpath_len + 1);
}

static int twab_astribank_count(int count, void *context)
{
	debug(TWINSTAR, "total Astribanks = %d\n", count);
	return 0;
}

struct twinstar_astribank_notify_ops twab_ops = {
	.astribank_notify	= twab_astribank_notify,
	.astribank_count	= twab_astribank_count,
};

static void twinstar_corosync_finalize(GObject *object)
{
	TwinstarCorosync	*twcorosync;
	TwinstarCorosyncPrivate	*priv;

	debug(TWINSTAR, "\n");
	g_return_if_fail(object != NULL);
	g_return_if_fail(TWINSTAR_IS_COROSYNC(object));
	twcorosync = TWINSTAR_COROSYNC(object);
	g_return_if_fail(twcorosync->priv != NULL);
	priv = twcorosync->priv;
	if (priv->sync_timer) {
		g_source_remove(priv->sync_timer);
		priv->sync_timer = 0;
	}
	if (priv->twlicense != NULL)
		twinstar_license_destroy(priv->twlicense);
	if (priv->twdbus != NULL)
		g_object_unref(priv->twdbus);
	if (priv->twgcpg != NULL)
		g_object_unref(priv->twgcpg);
	if (priv->twab != NULL)
		twinstar_astribanks_destroy(priv->twab);
	G_OBJECT_CLASS(twinstar_corosync_parent_class)->finalize(object);
}

static gboolean sync_timer_callback(gpointer user_data)
{
	TwinstarCorosync *twcorosync = user_data;
	TwinstarCorosyncPrivate	*priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	TwinstarGCpg *twgcpg = priv->twgcpg;
	TWINSTAR_MESSAGE_STRUCT(sync) *msg;
	int ret;

	/* Only master send "sync" packets */
	if (priv->current_master.nodeid != priv->local.nodeid)
		return TRUE;
	priv->local.sync_counter += 1 + (priv->local.priority * SYNC_PRIORITY_FACTOR);
	debug(TWINSTAR, "local.sync_counter = %u\n", priv->local.sync_counter);
	msg = ALLOCA_TWINSTAR_MESSAGE('s', sync, 0);
	if (!msg)
		return -ENOMEM;
	msg->payload.node = priv->local;
	ret = twcpg_glib_send_message(twgcpg, msg, sizeof(*msg));
	return (ret) ? TRUE : FALSE;
}

TwinstarCorosync *twinstar_corosync_new(GMainContext *mainctx)
{
	TwinstarCorosync	*twcorosync;
	TwinstarCorosyncPrivate *priv;

	/* Allocate */
	twcorosync = TWINSTAR_COROSYNC(g_object_new(TWINSTAR_TYPE_COROSYNC, NULL));
	priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);

	/* CPG initialization */
	priv->twgcpg = twcpg_glib_new("twinstar", &cpg_ops, twcorosync);
	if (!priv->twgcpg) {
		err("Failed twcpg_glib_new()\n");
		goto err;
	}
	if (!twcpg_glib_register(priv->twgcpg, mainctx)) {
		err("Failed twcpg_glib_register()\n");
		goto err;
	}
	priv->local.nodeid = twinstar_corosync_local_nodeid(twcorosync);
	debug(TWINSTAR, "local.nodeid=%u\n", priv->local.nodeid);

	/* Astribanks initialization */
	priv->twab = twinstar_astribanks_new(&twab_ops, twcorosync);
	if (!priv->twab) {
		err("Failed twinstar_astribanks_new()\n");
		goto err;
	}

	/* DBUS initialization */
	priv->twdbus = twinstar_dbus_new(priv->local.nodeid, &dbus_ops, twcorosync);
	if (!priv->twdbus) {
		err("Failed twinstar_dbus_new()\n");
		goto err;
	}

	/* License initialization */
	priv->twlicense = twinstar_license_new(XLIC_LICENSE_FILE, LICENSE_PRODUCT,
		&license_notify_ops, twcorosync);
	if (!priv->twlicense) {
		err("Failed twinstar_license_new()\n");
		goto err;
	}
	twinstar_license_validate_local(priv->twlicense);

	if (twinstar_get_preferred_master(priv->twdbus))
		priv->local.priority = 1;
	else
		priv->local.priority = 0;

	/* DBUS registration */
	if (!twinstar_dbus_register(priv->twdbus)) {
		err("Failed twinstar_dbus_register()\n");
		goto err;
	}
	twinstar_corosync_set_agents(twcorosync, "/usr/share/twinstar/run.d");
	priv->sync_timer = g_timeout_add_seconds(SYNC_INTERVAL, sync_timer_callback, twcorosync);
	return twcorosync;
err:
	if (twcorosync)
		g_object_unref(twcorosync);
	return NULL;
}

static void twinstar_corosync_init(TwinstarCorosync *twcorosync)
{
	TwinstarCorosyncPrivate *priv;
	debug(TWINSTAR, "\n");
	priv = TWINSTAR_COROSYNC_GET_PRIVATE(twcorosync);
	priv->election_timer = 0;
	priv->in_transition = FALSE;
	clear_node(&priv->current_master);
	priv->local.priority = NO_PRIORITY;
	twcorosync->priv = priv;
}

GQuark twinstar_corosync_error_quark(void)
{
	static GQuark ret = 0;
	if (ret == 0)
		ret = g_quark_from_static_string("twinstar_corosync_error");
	return ret;
}

#define ENUM_ENTRY(NAME, DESC) { NAME, "" #NAME "", DESC }

GType twinstar_corosync_error_get_type(void)
{
	static GType etype = 0;

	if (etype == 0) {
		static const GEnumValue values[] = {
			ENUM_ENTRY(TWINSTAR_COROSYNC_ERROR_GENERAL, "GeneralError"),
			{ 0, 0, 0 }
		};
		g_assert(TWINSTAR_COROSYNC_NUM_ERRORS == G_N_ELEMENTS(values) - 1);
		etype = g_enum_register_static("TwinstarCorosyncError", values);
	}
	return etype;
}

static void twinstar_corosync_class_init(TwinstarCorosyncClass *klass)
{
	GObjectClass	*object_class = G_OBJECT_CLASS(klass);

	debug(TWINSTAR, "\n");
	object_class->finalize = twinstar_corosync_finalize;
	//object_class->get_property = twinstar_corosync_get_property;
	//object_class->set_property = twinstar_corosync_set_property;
	g_type_class_add_private(klass, sizeof(TwinstarCorosyncPrivate));
}
