#include <dirent.h>
#include <errno.h>
#include <glib.h>
#include <glib-object.h>
#include <glob.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include "debug.h"
#include "twinstar-run.h"

G_DEFINE_TYPE(TwinstarRun, twinstar_run, G_TYPE_OBJECT);

#define TWINSTAR_RUN_GET_PRIVATE(o) (G_TYPE_INSTANCE_GET_PRIVATE((o), TWINSTAR_TYPE_RUN, TwinstarRunPrivate))

#define	RUN_DIR	"/usr/share/twinstar/run.d"

#define	RUN_GLOB_PATTERN	"*.run"
#define	RUN_CHILD_TIMEOUT	(240 * 1000)	/* msec */

struct TwinstarRunPrivate {
	const struct twinstar_run_notify_ops *notifications;
	char *run_dir;
	glob_t glob_space;
	void *context;
	int glob_idx;
	int child_pid;
	int child_watch_id;
	guint child_timer_id;
	gboolean is_running;
	gboolean is_master;
	gboolean errors;
	gboolean pending_transition;
	gboolean pending_transition_is_master;
};

static int run_next(TwinstarRun *twrun);

static void run_steps_done(TwinstarRun *twrun)
{
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);

	debug(RUN, "Finished (%s)\n", (priv->is_master) ? "master" : "slave");
	if (priv->child_watch_id) {
		err("a child watch active (removing) (pid=%d)\n", priv->child_pid);
		g_source_remove(priv->child_watch_id);
	}
	if (priv->child_timer_id) {
		err("a child timer active (removing) (pid=%d)\n", priv->child_pid);
		g_source_remove(priv->child_timer_id);
	}
	priv->glob_idx = 0;
	priv->is_running = 0;
	globfree(&priv->glob_space);
	priv->glob_space.gl_pathc = 0;
	priv->glob_space.gl_pathv = NULL;
	priv->glob_space.gl_offs = 0;
	if (priv->notifications && priv->notifications->run_steps_done)
		priv->notifications->run_steps_done(priv->context, priv->errors);
	if (priv->pending_transition) {
		priv->pending_transition = 0;
		info("Run pending transition (is_master=%d)\n",
			priv->pending_transition_is_master);
		twinstar_run_start(twrun, priv->pending_transition_is_master);
	}
}

static void run_step_advance(TwinstarRun *twrun)
{
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);

	debug(RUN, "idx=%d\n", priv->glob_idx);
	if (priv->is_running) {
		priv->glob_idx += (priv->is_master) ? +1 : -1;
		run_next(twrun);
	} else
		run_steps_done(twrun);
}

static void child_watch_callback(GPid pid, gint status, gpointer data)
{
	TwinstarRun *twrun = data;
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);
	const char *path;

	debug(RUN, "child_watch_callback(pid=%d, status=%d)\n", pid, status);
	if (priv->child_timer_id) {
		g_source_remove(priv->child_timer_id);
		priv->child_timer_id = 0;
	}
	g_spawn_close_pid(pid);
	if (pid != priv->child_pid) {
		err("%s: terminated pid (%d) != last run pid (%d)\n",
			__func__, pid, priv->child_pid);
	}
	if (status != 0)
		priv->errors++;
	if (! priv->glob_space.gl_pathv) {
		err("%s(%d, %d) -- empty glob_space -- ignore\n", __func__, pid, status);
		return;
	}
	path = priv->glob_space.gl_pathv[priv->glob_idx];
	if (priv->notifications && priv->notifications->run_step_end)
		priv->notifications->run_step_end(priv->context, path, priv->is_master, pid, status);
	run_step_advance(twrun);
}

static gboolean child_timeout(gpointer user_data)
{
	TwinstarRun *twrun = user_data;
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);

	priv->child_timer_id = 0;
	if (priv->child_pid) {
		err("child_timeout(pid=%d)\n", priv->child_pid);
		kill(priv->child_pid, SIGTERM);
	}
	return FALSE;	/* oneshot */
}

static int run_next(TwinstarRun *twrun)
{
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);
	int idx = priv->glob_idx;
	char *path;
	gchar *argv[3];
	pid_t pid;
	gboolean ret;
	GError *glib_error = NULL;
	gboolean done_all = FALSE;
	guint child_max_msec = RUN_CHILD_TIMEOUT;

	if (priv->is_master && idx >= priv->glob_space.gl_pathc)
		done_all = TRUE;
	if (!priv->is_master && idx < 0)
		done_all = TRUE;
	if (done_all) {
		run_steps_done(twrun);
		return 0;
	}
	path = priv->glob_space.gl_pathv[idx];
	debug(RUN, "running: %s (%d)\n", path, priv->is_master);
	argv[0] = path;
	argv[1] = (priv->is_master) ? "start" : "stop";
	argv[2] = NULL;
	ret = g_spawn_async(".", argv, NULL,  G_SPAWN_DO_NOT_REAP_CHILD, NULL, NULL, &pid, &glib_error);
	if (!ret) {
		err("%s\n", glib_error->message);
		g_error_free(glib_error);
		priv->child_pid = 0;
		priv->child_watch_id = 0;
		child_max_msec = 0;
		run_step_advance(twrun);
		return 0;
	}
	priv->child_pid = pid;
	priv->child_watch_id = g_child_watch_add(priv->child_pid, child_watch_callback, twrun);
	if (priv->notifications && priv->notifications->run_step_start)
		priv->notifications->run_step_start(priv->context, path, priv->is_master, pid);
	priv->child_timer_id = g_timeout_add(child_max_msec, child_timeout, twrun);
	return 0;
}

gboolean twinstar_run_start(TwinstarRun *twrun, gboolean is_master)
{
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);
	char path[PATH_MAX];
	int ret;

	debug(RUN, "\n");
	if (priv->is_running) {
		info("Already running -- remember pending transition (is_master=%d)\n",
			is_master);
		priv->pending_transition_is_master = is_master;
		priv->pending_transition = 1;
		return TRUE;
	}
	priv->is_running = 1;
	priv->errors = 0;
	priv->is_master = is_master;
	snprintf(path, sizeof(path), "%s/%s", priv->run_dir, RUN_GLOB_PATTERN);
	ret = glob(path, GLOB_ERR, NULL, &priv->glob_space);
	if (ret) {
		if (ret == GLOB_NOMATCH) {
			debug(RUN, "No match (%s)\n", path);
		} else {
			err("Failed glob(%s): %d\n", path, ret);
		}
	} else {
		priv->glob_idx = (is_master) ? 0 : priv->glob_space.gl_pathc - 1;
		run_next(twrun);
	}
	return TRUE;
}

TwinstarRun *twinstar_run_new(const char *run_dir,
	const struct twinstar_run_notify_ops *notifications,
	void *context)
{
	TwinstarRun *twrun;
	TwinstarRunPrivate *priv;

	debug(RUN, "\n");
	/* Allocate */
	twrun = TWINSTAR_RUN(g_object_new(TWINSTAR_TYPE_RUN, NULL));
	priv = TWINSTAR_RUN_GET_PRIVATE(twrun);

	/* Initialization */
	priv->run_dir = strdup(run_dir);
	priv->context = context;
	priv->notifications = notifications;
	return twrun;
}

static void twinstar_run_init(TwinstarRun *twrun)
{
	TwinstarRunPrivate *priv = TWINSTAR_RUN_GET_PRIVATE(twrun);

	debug(RUN, "\n");
	priv->notifications = NULL;
	priv->run_dir = NULL;
	priv->glob_space.gl_pathc = 0;
	priv->glob_space.gl_pathv = NULL;
	priv->glob_space.gl_offs = 0;
	priv->glob_idx = 0;
	priv->child_watch_id = 0;
	priv->child_timer_id = 0;
	priv->child_pid = 0;
	priv->is_running = 0;
	priv->is_master = 0;
	priv->context = NULL;
	priv->pending_transition = 0;
}

static void twinstar_run_finalize(GObject *object)
{
	TwinstarRun *twrun;
	TwinstarRunPrivate *priv;

	debug(RUN, "\n");
	g_return_if_fail(object != NULL);
	g_return_if_fail(TWINSTAR_IS_RUN(object));
	twrun = TWINSTAR_RUN(object);
	priv = TWINSTAR_RUN_GET_PRIVATE(twrun);
	g_return_if_fail(priv != NULL);
	if (priv->run_dir != NULL)
		g_free(priv->run_dir);
	G_OBJECT_CLASS(twinstar_run_parent_class)->finalize(object);
}

static void twinstar_run_class_init(TwinstarRunClass *klass)
{
	GObjectClass *object_class = G_OBJECT_CLASS(klass);

	debug(RUN, "\n");
	object_class->finalize = twinstar_run_finalize;
	//object_class->get_property = twinstar_run_get_property;
	//object_class->set_property = twinstar_run_set_property;
	g_type_class_add_private(klass, sizeof(TwinstarRunPrivate));
}

